// EE231002 Lab07. Magic Square
// 108061112, 林靖
// Nov. 2, 2019

#include <stdio.h>			   // Standard input and output library

#define TRUE 1				   // Denote Boolean data type
#define FALSE 0
#define SIZE 4				   // Size of magic square
#define UNFILLED 0			   // 0 represents blank space to be filled in

typedef short BOOL;			   // Set up Boolean data type with short basic type

int g_count;				   // Number of legal magic square found
int g_magic[SIZE][SIZE];	   // Store promising magic square
int g_target_sum;			   // Each row, column, diagonal add up to the same

void solve_unfilled_or_print(void);		  // Backtracking recursion function
BOOL sums_of_rows_are_legal(void);		  // If rows all sum up to the target
BOOL sums_of_columns_are_legal(void);	  // If columns all sum up to the target
BOOL sums_of_diagonals_are_legal(void);	  // If diagonals sum up to the target
void print_out_magic_square(void);		  // Print out the legal one we found


int main(void)							  // called at program start up
{
	int i, j;							  // For the loop reading inputs

	for (i = 0; i < SIZE; i++) {
		for (j = 0; j < SIZE; j++) {
			scanf("%d", &g_magic[i][j]);				// Read and store inputs
		}
	}
	g_target_sum = SIZE * (SIZE * SIZE + 1) / 2;		// Evaluate target sum
	solve_unfilled_or_print();							// Find the solutions
	printf("Number of solutions found: %d\n", g_count);	// Show total number
	return 0;											// Normal termination
}


/******************************************************************************
 * Find the next unfilled blank space and guess a non-repetitive integer to   *
 * fill in.	When all the spaces are filled, check if the sums of rows,		  *
 * columns, and diagonals all add up to the target sum.	If all the sums are	  *
 * legal, print out this magic square, and then backtrack the next one.		  *
 ******************************************************************************/
void solve_unfilled_or_print(void)
{
	BOOL found_unfilled;	// Check if there exists any blank space
	BOOL found_repetitive;	// Check if the integer has already been used
	int row, column;		// Keep track of indices in each recursion stack
	int i, j;				// Temp var for loop checking repetitive integers
	int guess;				// Tentative guess to fill into the blank spaces

	found_unfilled = FALSE;	// Initialize to check if there exists a blank space
	for (row = 0; !found_unfilled && row < SIZE; row++) {
		for (column = 0; !found_unfilled && column < SIZE; column++) {
			if (g_magic[row][column] == UNFILLED) {
				found_unfilled = TRUE;	// Flag that a blank space exists if we
			}							// find any.
		}
	}
	row--;							  // Fix off-by-one error after looping over
	column--;

	if (found_unfilled) {				// If there exists a blank space
		for (guess = SIZE * SIZE; guess > 0; guess--) {	// Guess a tentative num
			found_repetitive = FALSE;	// Initialize to check if it's been used
			for (i = 0; !found_repetitive && i < SIZE; i++) {
				for (j = 0; !found_repetitive && j < SIZE; j++) {
					if (g_magic[i][j] == guess) {
						found_repetitive = TRUE; // Flag that the tentative num
					}							 // has already been used.
				}
			}
			if (!found_repetitive) {	// If the tentative num hasn't been used
				g_magic[row][column] = guess;	 // Fill it into the blank space
				solve_unfilled_or_print();		 // Try to fill next blank space
				g_magic[row][column] = UNFILLED; // Undo the action in line 76
			}
		}
	} else {									 // If all the space are filled
//		if (sums_of_rows_are_legal() &&			 // Check rows added up
//			sums_of_columns_are_legal() &&		 // Check columns added up
//			sums_of_diagonals_are_legal()) {	 // Check diagonals added up
			print_out_magic_square();			 // Found a legal one so print
//		}
	}

	return;										// Indicate function termination
}


BOOL sums_of_rows_are_legal(void)		 // If rows all add up to the target
{
	int i, j;							 // Temp var for the loop checking sum
	int sum;							 // Store cumulative results of each row
	BOOL both_equal = TRUE;				 // If rows all add up to the target

	for (i = 0; both_equal && i < SIZE; i++) {	 // Check sum in each of rows
		sum = 0;		// Initialize to chech if this row adds up to the target
		for (j = 0; j < SIZE; j++) {
			sum += g_magic[i][j];				 // Add and accumulate the value
		}
		if (sum != g_target_sum) {	  // If sum of this row inequal to target,
			both_equal = FALSE;		  // flag that this magic square is illegal.
		}
	}

	return both_equal;			   // Return false if any sum of rows is illegal
}


BOOL sums_of_columns_are_legal(void)	  // If columns all add up to the target
{	
	int i, j;						// Temp var for the loop checking sum
	int sum;						// Store cumulative results of each column
	BOOL both_equal = TRUE;			// Check if columns all add up to the target

	for (j = 0; both_equal && j < SIZE; j++) {	 // Check sum in each of columns
		sum = 0;	 // Initialize to check if this column adds up to the target
		for (i = 0; i < SIZE; i++) {
			sum += g_magic[i][j];				 // Add and accumulate the value
		}
		if (sum != g_target_sum) {	// If sum of this column inequal to target,
			both_equal = FALSE;		// flag that this magic square is illegal.
		}
	}

	return both_equal;			// Retuen false if any sum of columns is illegal
}


BOOL sums_of_diagonals_are_legal(void)  // If diagonals all add up to the target
{
	int i, j;					   // Temp var for the loop checking sum
	int sum;					   // Store cumulative results of each diagonal
	BOOL both_equal = TRUE;		   // Check if diagovals add up to the target

	sum = 0; // Initialize to check if the main diagonal adds up to the target
	for (i = 0, j = 0; i < SIZE; i++, j++) {
		sum += g_magic[i][j];					 // Add and accumulate the value
	}
	if (sum != g_target_sum) {	   // If sum of main diagonal inequal to target,
		both_equal = FALSE;		   // flag that this magic square is illegal.
	}

	sum = 0;   // Initialize to chech is the anti-diagonal adds up to the target
	for (i = 0, j = SIZE - 1; i < SIZE; i++, j--) {
		sum += g_magic[i][j];					 // Add and accumulate the value
	}
	if (sum != g_target_sum) {	   // If sum of anti-diagonal inequal to target,
		both_equal = FALSE;		   // flag that this magic square is illegal.
	}

	return both_equal;					   // Return false if any sum is illegal
}


void print_out_magic_square(void)  // Print out when a legal one is found
{
	int i, j;					   // Temp var for the loop printing out

	printf("Solution %d:\n", ++g_count);   // Num of legal magic squares found
	for (i = 0; i < SIZE; i++) {
		for (j = 0; j < SIZE; j++) {
			printf("%4d", g_magic[i][j]);  // The integer in the magic square
		}
		printf("\n");	// (new line)
	}
	printf("\n");		// (new line)

	return;				// Indicates function termination
}
